from typing import Type

import pytest
import requests.exceptions
from requests_mock import Mocker as RequestsMocker

from bioimageio.spec._internal.validation_context import ValidationContext


@pytest.mark.parametrize(
    "url",
    [
        "https://example.com",
        "https://colab.research.google.com/github/bioimage-io/spec-bioimage-io/blob/main/example/load_model_and_create_your_own.ipynb",
        "https://www.kaggle.com/c/data-science-bowl-2018",
    ],
)
def test_httpurl_valid(url: str):
    from bioimageio.spec._internal.url import HttpUrl

    with ValidationContext(perform_io_checks=True):
        assert HttpUrl(url).exists(), url


@pytest.mark.parametrize(
    "text,status_code",
    [
        ("OK", 200),
        ("found", 302),
        ("redirected", 301),
        ("redirected", 303),
        ("redirected", 308),
        ("let's ignore this I guess??", 405),
    ],
)
def test_httpurl_mock_valid(text: str, status_code: int, requests_mock: RequestsMocker):
    from bioimageio.spec._internal.url import HttpUrl

    url = "https://mock_example.com"
    _ = requests_mock.get(url, text=text, status_code=status_code)
    assert HttpUrl(url).exists()


@pytest.mark.parametrize(
    "text,status_code",
    [
        ("forbidden", 403),
        ("Not found", 404),
        ("just wrong", 199),
    ],
)
def test_httpurl_mock_invalid(
    text: str, status_code: int, requests_mock: RequestsMocker
):
    from bioimageio.spec._internal.url import HttpUrl

    url = "https://mock_example.com"
    _ = requests_mock.head(url, text=text, status_code=status_code)
    _ = requests_mock.get(url, text=text, status_code=status_code)
    with ValidationContext(perform_io_checks=True):
        with pytest.raises(ValueError):
            _ = HttpUrl(url)

    with ValidationContext(perform_io_checks=False):
        assert not HttpUrl(url).exists()


@pytest.mark.parametrize(
    "exc",
    [
        requests.exceptions.InvalidURL,
    ],
)
def test_httpurl_mock_exc(exc: Type[Exception], requests_mock: RequestsMocker):
    from bioimageio.spec._internal.url import HttpUrl

    url = "https://mock_example.com"
    _ = requests_mock.head(url, exc=exc)
    with ValidationContext(perform_io_checks=True):
        with pytest.raises(ValueError):
            _ = HttpUrl(url)

    with ValidationContext(perform_io_checks=False):
        assert not HttpUrl(url).exists()


@pytest.mark.parametrize(
    "url,io_check",
    [
        ("https://example.invalid", True),
        ("https://example.invalid", False),
    ],
)
def test_httpurl_nonexisting(url: str, io_check: bool):
    from bioimageio.spec._internal.url import HttpUrl

    with ValidationContext(perform_io_checks=io_check):
        assert HttpUrl(url).exists(), url


@pytest.mark.parametrize(
    "url",
    [
        "invalid-url",
    ],
)
def test_httpurl_invalid(url: str):
    from bioimageio.spec._internal.url import HttpUrl

    with pytest.raises(ValueError), ValidationContext(perform_io_checks=False):
        _ = HttpUrl(url)



from datetime import datetime

import pytest

from bioimageio.spec._internal.validation_context import ValidationContext


def test_is_valid_yaml_value():
    from bioimageio.spec._internal.field_validation import is_valid_yaml_value

    class Obj:
        pass

    invalid = {"a": 1, "b": ["1", 2, Obj()]}

    assert not is_valid_yaml_value(invalid)

    valid = {"a": 1, "b": ["1", 2], ("c", 1): ("a", "b", datetime.now())}
    assert is_valid_yaml_value(valid)


def test_validate_unique_entries():
    from bioimageio.spec._internal.field_validation import validate_unique_entries

    with pytest.raises(ValueError):
        _ = validate_unique_entries(["a", "a"])

    ret = validate_unique_entries(["a", "b"])
    assert ret == ["a", "b"]


def test_validate_gh_user():
    from bioimageio.spec._internal.field_validation import validate_gh_user

    with pytest.raises(ValueError), ValidationContext(perform_io_checks=True):
        _ = validate_gh_user("arratemunoz")

    cpape = validate_gh_user("Constantin Pape", hotfix_known_errorenous_names=True)
    assert cpape == "constantinpape"

    fynnbe = validate_gh_user("fynnbe")
    assert fynnbe == "fynnbe"



import sys
from pathlib import Path

from typing_extensions import assert_never

from bioimageio.spec import SpecificResourceDescr
from scripts.generate_json_schemas import MAJOR_MINOR_VERSION

from . import generate_docs

html_result: "str | Exception" = generate_docs(
    raw_type=SpecificResourceDescr, root_path=["Delivery"]
)
if isinstance(html_result, Exception):
    print(f"Could not generate docs: {html_result}", file=sys.stderr)
    exit(1)
elif isinstance(html_result, str):
    docs_output_path = (
        Path(__file__).parent.parent.parent
        / f"dist/interactive_docs_{MAJOR_MINOR_VERSION}.html"
    )
    docs_output_path.parent.mkdir(parents=True, exist_ok=True)
    print(f"[INFO] Writing interactive docs to {docs_output_path}")
    with open(docs_output_path, "w", encoding="utf-8") as f:
        _ = f.write(html_result)
        print(f"Wrote {_} bytes to {docs_output_path}")
else:
    assert_never(html_result)



from typing import Any, List
from .hint import Hint, Unrecognized, Widget


def generate_docs(*, raw_type: Any, root_path: List[str]) -> "str | Exception":
    hint = Hint.parse(raw_hint=raw_type, parent_raw_hints=[])
    if isinstance(hint, (Exception, Unrecognized)):
        return Exception(f"Could not process {raw_type}: {hint}")

    root_hint_widget = hint.to_type_widget(root_path)
    assert isinstance(root_hint_widget, Widget), root_hint_widget

    return f"""
        <!doctype html>
        <html>
            <head>
                <meta charset="utf-8">
                <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/styles/school-book.min.css">
                <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/highlight.min.js"></script>
                <script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.9.0/languages/yaml.min.js"></script>
            </head>
            <script>
                document.addEventListener("DOMContentLoaded", () => {{
                    if(location.hash){{
                        console.log("Expanding details and jumping to hash");
                        const target = document.getElementById(location.hash.slice(1));
                        if(!target){{
                            return
                        }}
                        let parent = target.parentElement;
                        while(parent){{
                            if(parent.tagName == "DETAILS"){{
                                parent.open = true;
                            }}
                            parent = parent.parentElement;
                        }}
                        target.scrollIntoView({{
                            behavior: 'smooth',
                            block: 'start',
                        }});
                    }}
                }})
            </script>
            <style>
                {Widget.get_css()}
            </style>
            <body>
                {root_hint_widget.to_html()}
            </body>
            <script>
                hljs.highlightAll();
            </script>
        </html>

    """



import shutil
import sys
from pathlib import Path

from json_schema_for_humans.generate import (  # pyright: ignore[reportMissingTypeStubs]
    generate_from_filename,
)
from json_schema_for_humans.generation_configuration import (  # pyright: ignore[reportMissingTypeStubs]
    GenerationConfiguration,
)

if __name__ == "__main__":
    glob_pattern = "dist/bioimageio_schema_*.json"
    schema_paths = list(Path().glob(glob_pattern))
    if not schema_paths:
        raise FileNotFoundError(
            f"no json schema files found with pattern '{glob_pattern}'"
        )

    for schema_file_path in schema_paths:
        output_dir = Path() / f"dist/{schema_file_path.stem}"
        if output_dir.exists():
            shutil.rmtree(output_dir)

        output_dir.mkdir(parents=True)

        generate_from_filename(
            schema_file_name=schema_file_path,
            result_file_name=str(output_dir / "index.html"),
            config=GenerationConfiguration(
                link_to_reused_ref=False,
                deprecated_from_description=True,
                default_from_description=True,
                examples_as_yaml=True,
                expand_buttons=True,
                description_is_markdown=True,
                copy_css=True,
                copy_js=True,
            ),
        )

        print(f"Generated json schema docs at {output_dir}", file=sys.stderr)



import json
from typing import List, TypedDict

from ._description import ensure_description_is_dataset, ensure_description_is_model
from ._internal.io import (
    download,
    extract_file_name,
    get_sha256,
    identify_bioimageio_yaml_file_name,
    is_valid_bioimageio_yaml_name,
)
from ._internal.io_utils import load_array, save_array, write_yaml
from ._internal.utils import files

__all__ = [
    "download",
    "ensure_description_is_dataset",
    "ensure_description_is_model",
    "extract_file_name",
    "get_sha256",
    "get_spdx_licenses",
    "identify_bioimageio_yaml_file_name",
    "is_valid_bioimageio_yaml_name",
    "load_array",
    "save_array",
    "SpdxLicenseEntry",
    "SpdxLicenses",
    "write_yaml",
]


class SpdxLicenseEntry(TypedDict):
    isDeprecatedLicenseId: bool
    isKnownByZenodo: bool
    isOsiApproved: bool
    licenseId: str
    name: str
    reference: str


class SpdxLicenses(TypedDict):
    licenseListVersion: str
    licenses: List[SpdxLicenseEntry]
    releaseDate: str


def get_spdx_licenses() -> SpdxLicenses:
    """get details of the SPDX licenses known to bioimageio.spec"""
    with files("bioimageio.spec").joinpath("static/spdx_licenses.json").open(
        "r", encoding="utf-8"
    ) as f:
        return json.load(f)



from pathlib import Path
from typing import Any, ClassVar, Type

import pydantic
import zipp  # pyright: ignore[reportMissingTypeStubs]
from annotated_types import Predicate
from pydantic import RootModel, StringConstraints
from typing_extensions import Annotated

from .validated_string import ValidatedString

FileName = str
FilePath = Annotated[pydantic.FilePath, pydantic.Field(title="FilePath")]
AbsoluteDirectory = Annotated[
    pydantic.DirectoryPath,
    Predicate(Path.is_absolute),
    pydantic.Field(title="AbsoluteDirectory"),
]
AbsoluteFilePath = Annotated[
    pydantic.FilePath,
    Predicate(Path.is_absolute),
    pydantic.Field(title="AbsoluteFilePath"),
]

BIOIMAGEIO_YAML = "rdf.yaml"
ALTERNATIVE_BIOIMAGEIO_YAML_NAMES = ("bioimageio.yaml", "model.yaml")
ALL_BIOIMAGEIO_YAML_NAMES = (BIOIMAGEIO_YAML,) + ALTERNATIVE_BIOIMAGEIO_YAML_NAMES

ZipPath = zipp.Path  # not zipfile.Path due to https://bugs.python.org/issue40564


class Sha256(ValidatedString):
    """A SHA-256 hash value"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(
                strip_whitespace=True, to_lower=True, min_length=64, max_length=64
            ),
        ]
    ]



import json
from pathlib import Path

from setuptools import find_namespace_packages, setup

# Get the long description from the README file
ROOT_DIR = Path(__file__).parent.resolve()
long_description = (ROOT_DIR / "README.md").read_text(encoding="utf-8")
VERSION_FILE = ROOT_DIR / "bioimageio" / "spec" / "VERSION"
VERSION = json.loads(VERSION_FILE.read_text(encoding="utf-8"))["version"]

_ = setup(
    name="bioimageio.spec",
    version=VERSION,
    description="Parser and validator library for bioimage.io specifications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bioimage-io/spec-bioimage-io",
    author="bioimage.io Team",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    packages=find_namespace_packages(exclude=["tests"]),  # Required
    install_requires=[
        "annotated-types>=0.5.0,<1",
        "email_validator",
        "imageio",
        "loguru",
        "markdown",
        "numpy>=1.21",
        "packaging>=17.0",
        "pooch>=1.5,<2",
        "pydantic-settings>=2.5,<3",
        "pydantic>=2.7.0,<2.10",  # TODO: check and update pin after https://github.com/pydantic/pydantic/pull/11008 is released
        "python-dateutil",
        "requests",
        "rich",
        "ruyaml",
        "tifffile>=2020.7.4",
        "tqdm",
        "typing-extensions",
        "zipp",
    ],
    extras_require={
        "tests": (
            test_extras := [
                "black",
                "deepdiff",
                "jsonschema",
                "lxml",
                "pytest-cov",
                "pytest",
                "requests-mock",
            ]
        ),
        "dev": (
            test_extras
            + [
                # "json_schema_for_humans; python_version < 3.13",
                "jupyter-black",
                "jupyter",
                "pdoc",
                "pre-commit",
                "pyright==1.1.398",
                "ruff",
            ]
        ),
    },
    scripts=[],
    include_package_data=True,
    project_urls={
        "Bug Reports": "https://github.com/bioimage-io/spec-bioimage-io/issues",
        "Source": "https://github.com/bioimage-io/spec-bioimage-io",
    },
)



import datetime
import inspect
import json
import types
import typing
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import (
    Any,
    ClassVar,
    Dict,
    Final,
    ForwardRef,
    Literal,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
    final,
)
from xml.etree import ElementTree as et

import pydantic
import typing_extensions
import yaml
from annotated_types import Predicate
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from pydantic_core import PydanticUndefined, PydanticUndefinedType
from typing_extensions import List, TypeAlias, TypeAliasType, assert_never

from bioimageio.spec._internal.io import YamlValue, is_yaml_leaf_value, is_yaml_value
from bioimageio.spec._internal.validated_string import ValidatedString


def eprint(message: str):
    import sys

    print(message, file=sys.stderr)


class Unrecognized:
    raw_hint: Any

    def __init__(self, raw_hint: Any):
        self.raw_hint = raw_hint
        super().__init__()

    def __str__(self) -> str:
        return f"Unrecognized type hint: {self.raw_hint}"

    def with_context(self, message: str) -> "ParsingError":
        return ParsingError(message=message, cause=self)


class ParsingError(Exception):
    def __init__(
        self, message: str, cause: "ParsingError | Unrecognized | None" = None
    ) -> None:
        self.message = message
        self.cause = cause
        super().__init__()

    def _display(self, level: int) -> str:
        base_message = "  " * level + self.message
        if isinstance(self.cause, ParsingError):
            return base_message + ":\n" + self.cause._display(level + 1)
        elif isinstance(self.cause, Unrecognized):
            return base_message + ":\n" + ("  " * (level + 1)) + str(self.cause)
        elif self.cause is None:
            return base_message
        else:
            assert_never(self.cause)

    def __str__(self) -> str:
        return self._display(level=0)

    def with_context(self, message: str) -> "ParsingError":
        return ParsingError(message=message, cause=self)


def any_is_subclass(child: Any, parent: Type[Any]) -> bool:
    return inspect.isclass(child) and issubclass(child, parent)


@dataclass
class Example:
    value: YamlValue

    @classmethod
    def try_from_value(cls, val: Any) -> "Example | Exception":
        if is_yaml_value(val):
            return Example(value=val)
        if isinstance(val, BaseModel):
            yaml_value: YamlValue = cast(YamlValue, val.model_dump(mode="json"))
            return Example(value=yaml_value)
        try:
            # FIXME: stricter typing here?
            val_type: Any = type(val)
            adapter: Any = pydantic.TypeAdapter(val_type)
            dumped_value = json.loads(adapter.dump_json(val))
            return Example(value=dumped_value)
        except Exception as e:
            return Exception(f"Value {val} is not Json-serializable: {e}")

    def to_yaml_str(self) -> str:
        out = yaml.dump(self.value)
        yaml_end_of_file = (
            "\n...\n"  # FIXME: can't we just dump without the end-of-file marker?
        )
        if out.endswith(yaml_end_of_file):
            return out[: -len(yaml_end_of_file)]
        else:
            return out

    def to_json_str(self) -> str:
        return json.dumps(self.value, indent=4)


def get_field_annotation(field_info: FieldInfo) -> Any:
    if field_info.metadata:
        inner_annotation: Any = field_info.annotation  # FIXME?
        # FIXME: thisrequired python 3.11
        # return typing_extensions.Annotated[inner_annotation, *field_info.metadata]
        out = typing_extensions.Annotated[inner_annotation, "placeholder"]
        out.__metadata__ = (  # pyright: ignore [reportAttributeAccessIssue]
            field_info.metadata
        )
        return out
    else:
        return field_info.annotation


def field_has_default_value(info: FieldInfo) -> bool:
    if not isinstance(info.default, PydanticUndefinedType):
        return True
    if info.default_factory is not None:
        return True
    return False


class Hint(ABC):
    @staticmethod
    def get_subclasses() -> Sequence[Type["Hint"]]:
        return [
            YamlValueHint,
            RecursionHint,
            StringNodeHint,
            RootModelHint,
            AnnotatedHint,
            TypeAliasHint,
            DatetimeHint,
            DateHint,
            PathHint,
            EmailHint,
            UrlHint,
            MappingHint,
            LiteralHint,
            ModelHint,
            PrimitiveHint,
            NTuple,
            VarLenTuple,
            ListHint,
            UnionHint,
        ]

    @final
    @classmethod
    def parse(
        cls,
        *,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        # if raw_hint in cls.hint_cache:
        # return cls.hint_cache[raw_hint] #FIXME: maybe move this into the individual do_parse impls?
        hint: "Hint | Unrecognized | Exception" = Unrecognized(raw_hint=raw_hint)
        for subclass in Hint.get_subclasses():
            hint = subclass.do_parse(
                raw_hint, parent_raw_hints=parent_raw_hints, discriminator=discriminator
            )
            if isinstance(hint, ParsingError):
                return hint
            if isinstance(hint, Unrecognized):
                continue
            # cls.hint_cache[raw_hint] = hint
            break
        return hint

    def __init__(self) -> None:
        super().__init__()

    @classmethod
    @abstractmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        raise NotImplementedError

    @abstractmethod
    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        raise NotImplementedError

    @abstractmethod
    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        raise NotImplementedError

    @abstractmethod
    def get_example(self) -> "Example | Exception":
        raise NotImplementedError


class YamlValueHint(Hint):
    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        # FIXME: since the spec is yaml, "Any" mostly translates to YamlValue.... but is this always true?
        if raw_hint == typing.Any:
            return YamlValueHint()
        if raw_hint == "YamlValue":
            return YamlValueHint()
        if isinstance(raw_hint, ForwardRef) and raw_hint.__forward_arg__ == "YamlValue":
            return YamlValueHint()
        if inspect.isclass(raw_hint) and raw_hint.__name__ == "YamlValue":
            return YamlValueHint()
        if isinstance(raw_hint, TypeAliasType) and str(raw_hint) == "YamlValue":
            return YamlValueHint()
        return Unrecognized(raw_hint=raw_hint)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", css_classes=[FieldsWidget.FIELD_TYPE_CSS_CLASS], children=[
            InlinePre(text="YamlValue"),
            *extra
        ])
        # fmt: on

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)

    def get_example(self) -> "Example | Exception":
        return Example({"some_key": "some_value", "another_key": [123, 456]})


class RecursionHint(Hint):
    def __init__(self, raw_hint: Any) -> None:
        self.raw_hint = raw_hint
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        if raw_hint not in parent_raw_hints:
            return Unrecognized(raw_hint=raw_hint)
        return RecursionHint(raw_hint=raw_hint)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text=f"(Recursion to {self.raw_hint})"),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example("... RECURSE ...")

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        # fmt: off
        return Widget("span", css_classes=[FieldsWidget.FIELD_TYPE_CSS_CLASS], children=[
            self.short_description(extra=extra_summary)
        ])
        # fmt: on


class StringNodeHint(Hint):  # TODO: remove and add `ValidatedStringHint` instead
    def __init__(self, pattern: str) -> None:
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        if not inspect.isclass(raw_hint) or not any(
            klass.__name__ == "StringNode" for klass in raw_hint.__mro__
        ):
            return Unrecognized(raw_hint=raw_hint)
        return StringNodeHint(
            pattern=getattr(raw_hint, "_pattern")
        )  # FIXME: use types from spec

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="ISO 8601 datetime"),
            *extra
        ])
        # fmt: on

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)

    def get_example(self) -> "Example | Exception":
        return Example(
            "*NO EXAMPLE PROVIDED*"
        )  # FIXME: maybe have a type just for absent examples


class RootModelHint(Hint):
    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        from pydantic import RootModel

        if not inspect.isclass(raw_hint):
            return Unrecognized(raw_hint=raw_hint)
        if issubclass(raw_hint, RootModel):
            inner = raw_hint.model_fields["root"]
            inner_hint = get_field_annotation(inner)
        elif issubclass(raw_hint, ValidatedString):
            inner_hint = raw_hint.root_model
        else:
            return Unrecognized(raw_hint=raw_hint)
        # FIXME: this misses the extra restrictions that might be defined in `raw_hint`
        return Hint.parse(raw_hint=inner_hint, parent_raw_hints=parent_raw_hints)


class AnnotatedHint(Hint):
    CSS_CLASS: ClassVar[str] = "annotated_type"
    RESTRICTION_CSS_CLASS: ClassVar[str] = "typing_annotated_annotation"

    inner_hint: Hint
    restrictions: Sequence[Any]

    def __init__(self, inner_hint: Hint, restrictions: List[str]) -> None:
        self.inner_hint = inner_hint
        self.restrictions = restrictions
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        if raw_hint.__class__ != typing_extensions.Annotated[int, None].__class__:
            return Unrecognized(raw_hint)

        inner_hint: "Hint | Unrecognized | ParsingError"
        discri: Optional[pydantic.Discriminator] = discriminator
        for md in raw_hint.__metadata__:
            if not isinstance(md, pydantic.Discriminator):
                continue
            discri = md
            break
        inner_hint = Hint.parse(
            raw_hint=raw_hint.__args__[0],
            parent_raw_hints=[*parent_raw_hints, raw_hint],
            discriminator=discri,
        )
        if isinstance(inner_hint, (ParsingError, Unrecognized)):
            return inner_hint.with_context(f"Could not parse inner hint for {raw_hint}")
        if len(raw_hint.__metadata__) == 1 and isinstance(
            raw_hint.__metadata__[0], pydantic.Discriminator
        ):
            return inner_hint  # "discriminator" is onl helpful for the parser, not for the user
        metadata: List[str] = []
        for md in raw_hint.__metadata__:
            # pydantic.Discriminator and pydantic.PlainSerializer are not useful for consumers.
            # A FieldInfo can appear if in types like Annotated[int, pydantic.Field(...)]. It usually has nothing useful
            if isinstance(
                md,
                (
                    pydantic.Discriminator,
                    pydantic.PlainSerializer,
                    pydantic.fields.FieldInfo,
                ),
            ):
                continue
            # FIXME: use isinstance, maybe render it somehow?
            if type(md).__name__ == "AfterWarner":
                continue
            if "PydanticGeneralMetadata" in type(md).__name__:
                continue
            if isinstance(md, Predicate):
                if isinstance(md.func, types.LambdaType):
                    try:
                        metadata.append(inspect.getsource(md.func).strip())
                    except OSError:
                        eprint("WARNING: could not get lambda source")
                    continue
                metadata_str = md.func.__name__
                if md.func.__doc__:
                    metadata_str += f": {md.func.__doc__}"
                else:
                    eprint(
                        f"WARNING: no predicated docstring in {parent_raw_hints}.{raw_hint}"
                    )
                metadata.append(metadata_str)
                continue
            metadata.append(str(md))
        if metadata.__len__() == 0:
            return inner_hint
        if isinstance(inner_hint, AnnotatedHint):
            return AnnotatedHint(
                inner_hint=inner_hint.inner_hint,
                restrictions=metadata + list(inner_hint.restrictions),
            )
        return AnnotatedHint(inner_hint=inner_hint, restrictions=metadata)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        return self.inner_hint.short_description(
            extra=[
                WarningIconWidget(title="This field has further restrictions"),
                *extra,
            ]
        )

    def get_example(self) -> "Example | Exception":
        eprint("WARNING: Annotated type without a manually provided example")
        return self.inner_hint.get_example()

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        inner_widget = self.inner_hint.to_type_widget(
            path=path,
            extra_summary=[
                WarningIconWidget(title="This field has further restrictions"),
                *extra_summary,
            ],
        )

        # fmt: off
        restrictions_widget = Widget("div", children=[
            Widget("p", text="Restrictions:"),
            Widget("ul", css_classes=[self.RESTRICTION_CSS_CLASS], children=[
                Widget("li", text=str(annotation))
                for annotation in self.restrictions
            ]),
        ])
        # fmt: on

        if inner_widget.tag == "details":
            inner_widget.appendChildren([restrictions_widget])
            return inner_widget
        else:
            # fmt: off
            return Widget("details", children=[
                Widget("summary", children=[inner_widget]),
                restrictions_widget,
            ])
            # fmt: on


class TypeAliasHint(Hint):
    def __init__(self, name: str, inner: Hint) -> None:
        self.name = name
        self.inner = inner
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        if not isinstance(raw_hint, TypeAliasType):
            return Unrecognized(raw_hint=raw_hint)
        inner = Hint.parse(
            raw_hint=raw_hint.__value__,
            parent_raw_hints=[*parent_raw_hints, raw_hint],
            discriminator=discriminator,
        )
        if isinstance(inner, (Unrecognized, ParsingError)):
            return inner.with_context(
                f"Could not parse definition of alias {raw_hint.__name__}"
            )
        return TypeAliasHint(name=raw_hint.__name__, inner=inner)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text=f"{self.name}"),
            Widget("span", text=" (Alias)", style="font-style: italic; opacity: 0.6"),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        inner_example = self.inner.get_example()
        if isinstance(inner_example, Exception):
            return Exception(f"Could not get example for {self.name}: {inner_example}")
        return inner_example

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        # fmt: off
        return Widget("details", children=[
            Widget("summary", children=[
                self.short_description(extra=extra_summary)
            ]),
            self.inner.to_type_widget(path=path, extra_summary=[
                Widget("span", text=" (Aliased)", style="font-style: italic; opacity: 0.6"),
            ])
        ])
        # fmt: on


class DatetimeHint(Hint):
    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        if raw_hint != datetime.datetime:
            return Unrecognized(raw_hint=raw_hint)
        return DatetimeHint()

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="ISO 8601 datetime"),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example(datetime.datetime.now().isoformat())

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)


class DateHint(Hint):
    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        if raw_hint != datetime.date:
            return Unrecognized(raw_hint=raw_hint)
        return DateHint()

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="date (YYYY-MM-DD)"),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example("2024-12-31")

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)


class PathHint(Hint):
    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        from pathlib import Path, PurePath

        if raw_hint != Path and raw_hint != PurePath:
            return Unrecognized(raw_hint=raw_hint)
        return PathHint()

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="Path"),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example("/some/path")

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)


class EmailHint(Hint):
    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        from pydantic.networks import EmailStr

        if raw_hint != EmailStr:
            return Unrecognized(raw_hint=raw_hint)
        return EmailHint()

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="EmailStr"),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example("john.doe@example.com")

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)


class UrlHint(Hint):
    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        from pydantic import AnyUrl

        if not inspect.isclass(raw_hint) or not issubclass(raw_hint, AnyUrl):
            return Unrecognized(raw_hint=raw_hint)
        return UrlHint()

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="HttpUrl"),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example("https://example.com/some/path")

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)


class MappingHint(Hint):
    value_hint: Hint
    value_example: Example

    def __init__(self, value_hint: Hint, value_example: Example):
        self.value_hint = value_hint
        self.value_example = value_example
        super().__init__()

    @staticmethod
    def is_mapping_hint(
        hint: Any,
    ) -> bool:  # Can't use TypeGuard[typing.GenericAlias] in py 3.8
        if inspect.isclass(hint):
            return issubclass(hint, Mapping)
        if issubclass(type(hint), type(typing.Mapping[int, str])):
            return any_is_subclass(hint.__origin__, Mapping)
        return False

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "MappingHint | Unrecognized | ParsingError":
        if not cls.is_mapping_hint(raw_hint):
            return Unrecognized(raw_hint)
        type_args: Tuple[Type[Any], Type[Any]] = getattr(raw_hint, "__args__")
        key_type = type_args[0]
        if (
            key_type is not str
            and key_type is not int
            and key_type != typing.Union[int, str]
        ):
            return ParsingError(
                f"Mappings with keys that are not ints or strings is not supported yet: {raw_hint}"
            )
        value_hint = Hint.parse(
            raw_hint=raw_hint.__args__[1],
            parent_raw_hints=[*parent_raw_hints, raw_hint],
        )
        if isinstance(value_hint, (Unrecognized, ParsingError)):
            return value_hint.with_context(f"Could not parse value type of {raw_hint}")
        value_example = value_hint.get_example()
        if isinstance(value_example, Exception):
            return ParsingError(
                f"Could not get example for {value_hint}: {value_example}"
            )
        return MappingHint(value_hint=value_hint, value_example=value_example)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="{ str: "),
            self.value_hint.short_description(),
            InlinePre(text="}"),
            *extra,
        ])
        # fmt: on

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        field_name = "[...]"  # FIXME: clearer key type?
        return FieldsWidget(
            base_path=path,
            short_description=self.short_description(extra=extra_summary),
            fields=[
                FieldData(
                    name=field_name,
                    description=None,
                    type_widget=self.value_hint.to_type_widget(
                        path=[*path, field_name]
                    ),
                    example=self.value_example,
                )
            ],
        )

    def get_example(self) -> "Example":
        return Example({"some_key": self.value_example.value})


def literal_value_to_code(lit_value: "int | bool | float | str | None") -> str:
    return f"'{lit_value}'" if isinstance(lit_value, str) else str(lit_value)


class LiteralHint(Hint):
    LIMIT = 10

    def __init__(self, values: Sequence["int | float | bool | str | None"]):
        self.values = values
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "LiteralHint | Unrecognized | ParsingError":
        some_dummy_literal_hint = Literal["a"]
        if raw_hint.__class__ != some_dummy_literal_hint.__class__:
            return Unrecognized(raw_hint)
        assert all(
            isinstance(val, (int, float, str, type(None), bool))
            for val in raw_hint.__args__
        )
        return LiteralHint(values=raw_hint.__args__)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        widgets: List[Widget] = []

        if len(self.values) > LiteralHint.LIMIT:
            for part in self.values[0 : LiteralHint.LIMIT]:
                widgets.append(LiteralWidget(value=part))
                widgets.append(InlinePre(text=", "))
            widgets.append(
                InlinePre(text=f"({len(self.values) - LiteralHint.LIMIT} more)")
            )
            widgets.append(InlinePre(text=", "))
            widgets.append(LiteralWidget(value=self.values[-1]))
        else:
            for part_index, part in enumerate(self.values):
                widgets.append(LiteralWidget(value=part))
                if part_index < len(self.values) - 1:
                    widgets.append(InlinePre(text=", "))
        return Widget("span", children=[*widgets, *extra])

    def get_example(self) -> "Example | Exception":
        return Example(self.values[0])

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        if len(self.values) <= LiteralHint.LIMIT:
            # fmt: off
            return Widget("span", css_classes=[FieldsWidget.FIELD_TYPE_CSS_CLASS], children=[
                self.short_description(extra=extra_summary)
            ])
            # fmt: on

        # fmt: off
        return Widget("details", children=[
            Widget("summary", children=[
                self.short_description(extra=extra_summary)
            ]),
            Widget("table", children=[
                Widget("tbody", children=[
                    Widget("tr", children=[
                        Widget("td", css_classes=[FieldsWidget.FIELD_TYPE_CSS_CLASS], children=[
                            LiteralWidget(value=val)
                        ])
                    ])
                    for val in self.values
                ])
            ])
        ])
        # fmt: on


class ModelHint(Hint):
    def __init__(
        self,
        model: Type["BaseModel"],
        fields: Mapping[str, Tuple[Hint, Example]],
        discriminator: Optional[pydantic.Discriminator],
    ):
        self.model = model
        self.fields = fields
        self.discriminator = discriminator
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "Hint | Unrecognized | ParsingError":
        if not inspect.isclass(raw_hint) or not issubclass(raw_hint, BaseModel):
            return Unrecognized(raw_hint)

        fields: Dict[str, Tuple[Hint, Example]] = {}

        required_fields_first = sorted(
            raw_hint.model_fields.items(),
            key=lambda item: field_has_default_value(item[1]),
        )

        for field_name, field_info in required_fields_first:
            field_descriptor = f"{raw_hint.__name__}.{field_name}"
            field_hint = Hint.parse(
                raw_hint=get_field_annotation(field_info),
                # raw_hint=typing.get_type_hints(raw_hint, include_extras=True)[field_name],
                parent_raw_hints=[*parent_raw_hints, raw_hint],
                discriminator=None,  # discard discriminator as it only applies to the current ModelHint
            )

            if isinstance(field_hint, (ParsingError, Unrecognized)):
                return field_hint.with_context(
                    f"Could not parse type of field {field_descriptor}"
                )

            if field_info.examples is None or len(field_info.examples) == 0:
                raw_field_example = field_hint.get_example()
                if isinstance(raw_field_example, Exception):
                    return ParsingError(
                        f"Could not get example for {field_descriptor}: {raw_field_example}"
                    )
                raw_field_example = raw_field_example.value
            else:
                raw_field_example = field_info.examples[0]

            field_example = Example.try_from_value(raw_field_example)
            if isinstance(field_example, Exception):
                return ParsingError(
                    f"Could not get example for {field_descriptor}: {field_example}"
                )

            fields[field_name] = (field_hint, field_example)

        return ModelHint(model=raw_hint, fields=fields, discriminator=discriminator)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text=self.model.__module__ + "." + self.model.__qualname__),
            *extra
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example(
            {
                field_name: field_example.value
                for field_name, (_, field_example) in self.fields.items()
            }
        )

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        fields: List[FieldData] = []
        for field_name, (hint, example) in self.fields.items():
            field_info = self.model.model_fields[field_name]
            field_default = field_info.default
            if field_info.default_factory is not None:
                field_default = field_info.default_factory()
            if not isinstance(field_default, PydanticUndefinedType):
                field_default = Example.try_from_value(field_default)
                assert not isinstance(field_default, Exception)
            if self.discriminator and self.discriminator.discriminator == field_name:
                field_default = PydanticUndefined
            fields.append(
                FieldData(
                    name=field_name,
                    description=field_info.description,
                    example=example,
                    type_widget=hint.to_type_widget(path=[*path, field_name]),
                    default=field_default,
                )
            )

        return FieldsWidget(
            base_path=path,
            short_description=self.short_description(extra=extra_summary),
            fields=fields,
        )


PrimitiveType: TypeAlias = Union[
    Type[int], Type[float], Type[bool], Type[str], Type[None]
]


class PrimitiveHint(Hint):
    def __init__(self, hint_type: PrimitiveType):
        self.hint_type = hint_type
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "PrimitiveHint | Unrecognized | ParsingError":
        if raw_hint is None:
            raw_hint = type(None)
        if not inspect.isclass(raw_hint) or not issubclass(
            raw_hint, (int, float, bool, str, type(None))
        ):
            return Unrecognized(raw_hint=raw_hint)
        return PrimitiveHint(hint_type=raw_hint)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="null" if self.hint_type is type(None) else self.hint_type.__name__),
            *extra
        ])
        # fmt: on

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return self.short_description(extra=extra_summary)

    def get_example(self) -> Example:
        hint_type = self.hint_type
        if hint_type is int:
            return Example(123456)
        if hint_type is float:
            return Example(3.14)
        if hint_type is bool:
            return Example(True)
        if hint_type is str:
            return Example("some free-format string")
        if hint_type is type(None):
            return Example(None)
        return Example("--- NO EXAMPLES PROVIDED ---")


def is_tuple_hint(
    hint: Any,
) -> bool:  # FIXME: can't use TypeGuard[types.GenericAlias] in py 3.8
    if inspect.isclass(hint) and hint.__name__ == "tuple":
        return True
    if issubclass(type(hint), type(typing.Tuple[int, str])):
        return any_is_subclass(hint.__origin__, tuple)
    return False


class NTuple(Hint):
    """Represents tuple a type-hint with all items defined, like Tuple[int, str]"""

    def __init__(
        self,
        generic_args: Sequence[Tuple[Hint, Example]],
    ):
        self.generic_args = generic_args
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "NTuple | Unrecognized | ParsingError":
        if not is_tuple_hint(raw_hint) or (... in raw_hint.__args__):
            return Unrecognized(raw_hint)
        generic_args: List[Tuple[Hint, Example]] = []
        for arg_inx, arg in enumerate(raw_hint.__args__):
            hint = Hint.parse(
                raw_hint=arg, parent_raw_hints=[*parent_raw_hints, raw_hint]
            )
            if isinstance(hint, (Unrecognized, ParsingError)):
                return hint.with_context(
                    f"Could not parse {arg_inx}-th NTuple type argument"
                )
            example = hint.get_example()
            if isinstance(example, Exception):
                return ParsingError(
                    f"Could not get example for {arg_inx}-th NTuple type argument: {example}"
                )
            generic_args.append((hint, example))
        return NTuple(generic_args=generic_args)

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        children: List[Widget] = []
        for arg_idx, (arg, _example) in enumerate(self.generic_args):
            children.append(arg.short_description())
            if arg_idx < len(self.generic_args) - 1:
                children.append(InlinePre(text=", "))

        # fmt: off
        return Widget("span", children=[
            InlinePre(text="("),
            *children,
            InlinePre(text=")"),
            *extra,
        ])
        # fmt: on

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        return FieldsWidget(
            base_path=path,
            short_description=self.short_description(extra=extra_summary),
            fields=[
                FieldData(
                    name=str(idx),
                    description=None,
                    type_widget=hint.to_type_widget(path=[*path, str(idx)]),
                    example=example,
                )
                for idx, (hint, example) in enumerate(self.generic_args)
            ],
        )

    def get_example(self) -> "Example | Exception":
        return Example(list(arg[1].value for arg in self.generic_args))


class VarLenTuple(Hint):
    """Represents a type-hint like Tuple[T, ...]"""

    def __init__(self, element_type: Hint, element_example: Example):
        self.element_type = element_type
        self.element_example = element_example
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "VarLenTuple | Unrecognized | ParsingError":
        if not is_tuple_hint(raw_hint):
            return Unrecognized(raw_hint=raw_hint)
        if len(raw_hint.__args__) != 2:
            return Unrecognized(raw_hint=raw_hint)
        last_type_arg = raw_hint.__args__[-1]
        if last_type_arg.__name__ != "ellipsis":
            return Unrecognized(raw_hint=raw_hint)
        element_hint = Hint.parse(
            raw_hint=raw_hint.__args__[0],
            parent_raw_hints=[*parent_raw_hints, raw_hint],
        )
        if isinstance(element_hint, (Unrecognized, Exception)):
            return element_hint.with_context(
                f"Could not parse VarLenTuple element type: {element_hint}"
            )
        element_example = element_hint.get_example()
        if isinstance(element_example, Exception):
            return ParsingError(
                f"Could not get example for element type: {element_example}"
            )
        return VarLenTuple(element_type=element_hint, element_example=element_example)

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        field_name = "[...]"
        return FieldsWidget(
            base_path=path,
            fields=[
                FieldData(
                    name=field_name,
                    description=None,
                    type_widget=self.element_type.to_type_widget(
                        path=[*path, field_name]
                    ),
                    example=self.element_example,
                )
            ],
            short_description=self.short_description(extra=extra_summary),
        )

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="("),
            self.element_type.short_description(),
            InlinePre(text=", ...)"),
            *extra,
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example([self.element_example.value])


class ListHint(Hint):
    """Represents a list-hint like List[T]"""

    def __init__(self, element_type: Hint, element_example: Example):
        self.element_type = element_type
        self.element_example = element_example
        super().__init__()

    @staticmethod
    def is_list_hint(
        hint: Any,
    ) -> bool:  # FIXME: can't use TypeGuard[types.GenericAlias] in py 3.8:
        if inspect.isclass(hint):
            return issubclass(hint, Sequence)
        if issubclass(type(hint), type(typing.Sequence[int])):
            return any_is_subclass(hint.__origin__, Sequence)
        return False

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "ListHint | Unrecognized | ParsingError":
        if not cls.is_list_hint(raw_hint):
            return Unrecognized(raw_hint=raw_hint)
        element_hint = Hint.parse(
            raw_hint=raw_hint.__args__[0],
            parent_raw_hints=[*parent_raw_hints, raw_hint],
        )
        if isinstance(element_hint, (Unrecognized, ParsingError)):
            return element_hint.with_context("Could not parse List element type")
        element_example = element_hint.get_example()
        if isinstance(element_example, Exception):
            return ParsingError(
                f"Could not get example for List element type: {element_example}"
            )
        return ListHint(element_type=element_hint, element_example=element_example)

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        field_name = "[...]"
        return FieldsWidget(
            base_path=path,
            fields=[
                FieldData(
                    name=field_name,
                    description=None,
                    type_widget=self.element_type.to_type_widget(
                        path=[*path, field_name]
                    ),
                    example=self.element_example,
                )
            ],
            short_description=self.short_description(extra=extra_summary),
        )

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        # fmt: off
        return Widget("span", children=[
            InlinePre(text="List["),
            self.element_type.short_description(),
            InlinePre(text="]"),
            *extra,
        ])
        # fmt: on

    def get_example(self) -> "Example | Exception":
        return Example([self.element_example.value])


class UnionHint(Hint):
    def __init__(self, args: Sequence[Tuple[Hint, Example]]):
        self.variant_hints = args
        super().__init__()

    @classmethod
    def do_parse(
        cls,
        raw_hint: Any,
        parent_raw_hints: Sequence[Any],
        discriminator: Optional[pydantic.Discriminator] = None,
    ) -> "UnionHint | Unrecognized | ParsingError":
        some_dummy_union = Union[int, str]
        if raw_hint.__class__ != some_dummy_union.__class__:
            return Unrecognized(raw_hint=raw_hint)

        union_args: List[Tuple[Hint, Example]] = []
        for arg_idx, arg in enumerate(raw_hint.__args__):
            hint = Hint.parse(
                raw_hint=arg,
                parent_raw_hints=[*parent_raw_hints, raw_hint],
                discriminator=discriminator,
            )
            if isinstance(hint, (Unrecognized, ParsingError)):
                return hint.with_context(
                    f"Could not parse Union arg #{arg_idx} in {raw_hint}"
                )
            if isinstance(hint, UnionHint):
                union_args += list(hint.variant_hints)
            else:
                example = hint.get_example()
                if isinstance(example, Exception):
                    return ParsingError(
                        f"Could not get example for variant {hint}: {example}"
                    )
                union_args.append((hint, example))
        return UnionHint(args=union_args)

    def to_type_widget(
        self, path: List[str], extra_summary: Sequence["Widget"] = ()
    ) -> "Widget":
        variant_widgets: List[Widget] = []
        for variant_index, (variant_hint, variant_example) in enumerate(
            self.variant_hints
        ):
            variant_path = [
                *path,
                f"VARIANT_{variant_index}",
            ]  # FIXME: maybe variant name?
            variant_type_widget = variant_hint.to_type_widget(path=variant_path)

            # fmt: off
            variant_widgets.append(Widget("tr", children=[
                Widget("td", css_classes=[FieldsWidget.FIELD_TYPE_CSS_CLASS], children=[
                    variant_type_widget
                ]),
                Widget("td", css_classes=[FieldsWidget.EXAMPLE_FIELD_CSS_CLASS], children=[ExampleWidget(variant_example)])
            ]))
            # fmt: on

        # fmt: off
        variants_table = Widget("table", children=[
            Widget("thead", children=[
                Widget("tr", children=[
                    Widget("th", text="variant type", css_classes=[FieldsWidget.TYPE_TABLE_HEADER]),
                    Widget("th", text="example yaml", css_classes=[FieldsWidget.EXAMPLE_TABLE_HEADER]),
                ])
            ]),
            Widget("tbody", children=variant_widgets)
        ])
        #fmt: on


        #fmt: off
        return Widget("details", children=[
            Widget("summary", children=[
                self.short_description(extra=extra_summary),
                ColumnControls(root_element_id=variants_table.element_id),
            ]),
            variants_table
        ])
        # fmt: on

    def short_description(self, extra: Sequence["Widget"] = ()) -> "Widget":
        children: List[Widget] = []
        if len(extra) > 0:
            children.append(InlinePre(text="("))
        for arg_idx, (arg_hint, _) in enumerate(self.variant_hints):
            children.append(arg_hint.short_description())
            if arg_idx < len(self.variant_hints) - 1:
                children.append(InlinePre(text=" | "))
        if len(extra) > 0:
            children.append(InlinePre(text=")"))
        return Widget("span", children=[*children, *extra])

    def get_example(self) -> "Example":
        return self.variant_hints[0][1]


class Widget:
    id_counter: ClassVar[int] = 0

    tag: Final[str]
    element: Final[et.Element]
    subclasses: ClassVar[List[Type["Widget"]]] = []

    def __init__(
        self,
        tag: str,
        *,
        text: Optional[str] = None,
        children: Sequence["Widget"] = (),
        css_classes: Optional[List[str]] = None,
        title: str = "",
        style: str = "",
        element_id: str = "",
        attribs: Mapping[str, str] = {},
    ):
        self.tag = tag
        self.element = et.Element(tag)
        self.element_id = element_id or f"autoid_{Widget.id_counter}"
        Widget.id_counter += 1

        if text is not None:
            self.element.text = text
        if css_classes and len(css_classes) > 0:
            self.element.set("class", " ".join(css_classes))
        if title:
            self.element.set("title", title)
        if style:
            self.element.set("style", style)
        for key, val in attribs.items():
            self.element.set(key, val)
        self.element.set("id", self.element_id)
        super().__init__()

        self.appendChildren(children)

    def appendChildren(self, children: Sequence["Widget"]):
        for child in children:
            self.element.append(child.element)

    def create_child(self, tag: str, children: Sequence["Widget"] = ()) -> "Widget":
        out = Widget(tag=tag, children=children)
        self.appendChildren([out])
        return out

    def to_html(self) -> str:
        return et.tostring(self.element, method="html", encoding="unicode")

    @classmethod
    def get_css(cls) -> str:
        if cls != Widget:
            return ""

        theme_border = "solid 1px black"
        theme_spacing = "0.2em"
        theme_color_warning = "yellow"

        subclasses_css = "\n".join(klass.get_css() for klass in Widget.subclasses)

        return f"""
            {subclasses_css}

            table{{
                border-collapse: collapse;
            }}
            th{{
                background-color: #312f46;
                color: white;
            }}
            .{AnnotatedHint.RESTRICTION_CSS_CLASS}{{
                background-color: {theme_color_warning};
            }}
            td{{
                vertical-align: top;
            }}

            /*border*/
            td, th{{
                border-collapse: collapse;
            }}
            tr{{
                border-bottom: {theme_border};
            }}
            tr:last-child{{
                border-bottom: 0;
            }}

            th{{
                padding: {theme_spacing};
            }}
            td{{
                padding: 0;
            }}
            td:has(> table){{
                padding: 0;
            }}

            details{{
                padding: 0;
            }}
            details > table{{
                margin-left: 1em;
            }}
        """

    @classmethod
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        Widget.subclasses.append(cls)


class LiteralWidget(Widget):
    CSS_CLASS: ClassVar[str] = "literal_widget"

    @classmethod
    def get_css(cls) -> str:
        return f"""
            .{LiteralWidget.CSS_CLASS}{{
                color: #254604;
                font-family: monospace, monospace;
            }}
        """

    def __init__(self, *, value: "str | int | None | float"):
        super().__init__(
            "span", text=json.dumps(value), css_classes=[LiteralWidget.CSS_CLASS]
        )


class ExampleWidget(Widget):
    CSS_CLASS: ClassVar[str] = "example_widget"

    @classmethod
    def get_css(cls) -> str:
        return f"""
            .{ExampleWidget.CSS_CLASS}{{
                margin: 0;
            }}
        """

    def __init__(self, example: Example) -> None:
        text = example.to_yaml_str()
        if is_yaml_leaf_value(example.value):
            super().__init__(
                "pre",
                css_classes=[self.CSS_CLASS],
                children=[Widget("code", text=text, css_classes=["language-yaml"])],
            )
        else:
            super().__init__(
                "details",
                children=[
                    Widget("summary", text="example yaml"),
                    Widget(
                        "pre",
                        css_classes=[self.CSS_CLASS],
                        children=[
                            Widget("code", text=text, css_classes=["language-yaml"])
                        ],
                    ),
                ],
            )


class InlinePre(Widget):
    def __init__(
        self,
        *,
        text: Optional[str] = None,
        children: Sequence["Widget"] = (),
        css_classes: Optional[List[str]] = None,
    ):
        super().__init__(
            "span",
            text=text,
            children=children,
            css_classes=css_classes,
            style="font-family: monospace, monospace;",
        )


class WarningIconWidget(Widget):
    def __init__(self, title: str) -> None:
        super().__init__(
            "span",
            text="!",
            style="background-color: yellow; margin-right: 0.2em; margin-left: 0.2em; font-weight: bold;",
            title=title,
        )


class InfoIconWidget(Widget):
    CSS_CLASS: ClassVar[str] = "info_icon_widget"

    @classmethod
    def get_css(cls) -> str:
        return f"""
            .{InfoIconWidget.CSS_CLASS}{{
                opacity: 0.8;
                outline: solid 1px black;
                float: right;
                cursor: default;
            }}
        """

    def __init__(self, title: str) -> None:
        super().__init__(
            "span", text="?", css_classes=[InfoIconWidget.CSS_CLASS], title=title
        )


class OptMarkerWidget(Widget):
    CSS_CLASS: ClassVar[str] = "opt_marker_widget"

    @classmethod
    def get_css(cls) -> str:
        return f"""
            .{OptMarkerWidget.CSS_CLASS}{{
                background-color: yellow;
                color: black;
                font-weight: bold;
                padding: 1px;
                cursor: pointer;
                margin-right: 0.1em;
            }}
        """

    def __init__(self, default_value: Example):
        super().__init__(
            "span",
            text="opt",
            title=f"This field is optional and defaults to {default_value.to_json_str()}",
            css_classes=[OptMarkerWidget.CSS_CLASS],
        )


class CheckboxWidget(Widget):
    def __init__(
        self,
        *,
        text: Optional[str] = None,
        children: Sequence["Widget"] = (),
        on_click: str = "",
        css_classes: Optional[List[str]] = None,
        title: str = "",
        style: str = "",
    ):
        super().__init__("span")
        self._checkbox = Widget(
            "input",
            text=text,
            children=children,
            css_classes=css_classes,
            title=title,
            style=style,
        )
        self._checkbox.element.set("type", "checkbox")
        self._checkbox.element.set("checked", "checked")
        self.appendChildren([self._checkbox])
        if on_click:
            self.appendChildren(
                [
                    Widget(
                        "script",
                        text=f"""
                    document.getElementById("{self._checkbox.element_id}").addEventListener("click", (ev) => {{
                        {on_click}
                    }});
                """,
                    )
                ]
            )


class FragmentAnchorWidget(Widget):
    CSS_CLASS: Final[str] = "fragment_anchor_widget"

    @classmethod
    def get_css(cls) -> str:
        return f"""
            .{cls.CSS_CLASS}{{
                color: black;
                text-decoration: none;
                cursor: pointer;
            }}
            .{cls.CSS_CLASS}:hover{{
                text-decoration: underline;
            }}
        """

    def __init__(
        self,
        *,
        text: str = "",
        children: Sequence["Widget"] = (),
        path: List[str],
    ):
        fragment_contents = ".".join(path)
        super().__init__(
            "a",
            text=text,
            children=children,
            css_classes=[self.CSS_CLASS],
            title=fragment_contents,
            element_id=fragment_contents,
            attribs={"href": f"#{fragment_contents}"},
        )


class ColumnControls(Widget):
    CSS_CLASS: ClassVar[str] = "column_controls_widget"
    TYPES_CHECKBOX_CSS_CLASS: Final[str] = "types_checkbox"
    EXAMPLES_CHECKBOX_CSS_CLASS: Final[str] = "examples_checkbox"

    @classmethod
    def get_css(cls) -> str:
        return f"""
            .{ColumnControls.CSS_CLASS}{{
                font-weight: normal;
                background-color: rgba(0,0,0, 0.1);
            }}
        """

    def __init__(
        self,
        root_element_id: str,
        style: str = "float: right; margin-left: 4em; margin-right: 1em;",
    ):
        # fmt: off
        super().__init__(
            "span",
            style=style,
            css_classes=[ColumnControls.CSS_CLASS],
            children=[
                CheckboxWidget(
                    text="Ty",
                    css_classes=[self.TYPES_CHECKBOX_CSS_CLASS],
                    on_click=f"""
                        const table = document.querySelector("#{root_element_id}");
                        for(const ex of table.querySelectorAll(".{FieldsWidget.FIELD_TYPE_CSS_CLASS}, .{FieldsWidget.TYPE_TABLE_HEADER}")){{
                            ex.style.display = ev.target.checked ? "" : "none"
                        }}
                        for(const cb of table.querySelectorAll(".{self.TYPES_CHECKBOX_CSS_CLASS}")){{
                            cb.checked = ev.target.checked
                        }}
                    """,
                ),
                CheckboxWidget(
                    text="e.g.",
                    css_classes=[self.EXAMPLES_CHECKBOX_CSS_CLASS],
                    on_click=f"""
                        const table = document.querySelector("#{root_element_id}");
                        for(const ex of table.querySelectorAll(".{FieldsWidget.EXAMPLE_FIELD_CSS_CLASS}, .{FieldsWidget.EXAMPLE_TABLE_HEADER}")){{
                            ex.style.display = ev.target.checked ? "" : "none"
                        }}
                        for(const cb of table.querySelectorAll(".{self.EXAMPLES_CHECKBOX_CSS_CLASS}")){{
                            cb.checked = ev.target.checked
                        }}
                    """,
                ),
            ],
        )
        # fmt: on


@dataclass
class FieldData:
    name: str
    description: Optional[str]
    type_widget: Widget
    example: Optional[Example]
    default: "Example | PydanticUndefinedType" = PydanticUndefined


class FieldsWidget(Widget):
    FIELD_NAME_CSS_CLASS: ClassVar[str] = "field_name"
    FIELD_TYPE_CSS_CLASS: ClassVar[str] = "field_type"
    TYPE_TABLE_HEADER: ClassVar[str] = "type_header"
    EXAMPLE_TABLE_HEADER: ClassVar[str] = "example_header"
    EXAMPLE_FIELD_CSS_CLASS: ClassVar[str] = "example_field"
    TYPES_CHECKBOX_CSS_CLASS: ClassVar[str] = "types_checkbox"
    EXAMPLES_CHECKBOX_CSS_CLASS: ClassVar[str] = "examples_checkbox"

    @classmethod
    def get_css(cls) -> str:
        return f"""
            .{cls.FIELD_NAME_CSS_CLASS}{{
                background-color: white;
                font-weight: normal;
                padding: 0.3em;
            }}
            .{cls.FIELD_TYPE_CSS_CLASS}{{
                background-color: #84b4dd;
                font-weight: bold;
                padding: 0.3em;
                padding-right: 0;
                padding-bottom: 0;
            }}
            .{cls.EXAMPLE_FIELD_CSS_CLASS}{{
                background-color: #f6f5b2;
                font-weight: normal;
            }}
            .{cls.EXAMPLE_TABLE_HEADER}{{

            }}
        """

    def __init__(
        self,
        base_path: List[str],
        short_description: Widget,
        fields: Sequence[FieldData],
    ) -> None:
        # fmt: off
        fields_table = Widget("table", children=[
            Widget("thead", children=[
                Widget("tr", children=[
                    Widget("th", text="field name"),
                    Widget("th", text="field type", css_classes=[self.TYPE_TABLE_HEADER]),
                    Widget("th", text="example yaml", css_classes=[self.EXAMPLE_TABLE_HEADER]),
                ])
            ]),
            Widget("tbody", children=[
                Widget("tr", children=[
                    Widget("td", css_classes=[self.FIELD_NAME_CSS_CLASS], children=[
                        FragmentAnchorWidget(text=field.name, path=[*base_path, field.name]),
                        *([OptMarkerWidget(default_value=field.default)] if not isinstance(field.default, PydanticUndefinedType) else []),
                        *([InfoIconWidget(title=field.description)] if field.description else [])
                    ]),
                    Widget("td", css_classes=[self.FIELD_TYPE_CSS_CLASS], children=[field.type_widget]),
                    Widget("td", css_classes=[self.EXAMPLE_FIELD_CSS_CLASS], children=[
                        ExampleWidget(example=field.example) if field.example is not None else Widget("span", text="NO EXAMPLE")
                    ])
                ])
                for field in fields
            ]),
        ])
        # fmt: on

        # fmt: off
        super().__init__("details", children=[
            Widget("summary", children=[
                FragmentAnchorWidget(children=[short_description], path=base_path),
                ColumnControls(root_element_id=fields_table.element_id)
            ]),
            fields_table,
        ])
        # fmt: on



from __future__ import annotations

import hashlib
import io
import sys
import warnings
import zipfile
from abc import abstractmethod
from contextlib import nullcontext
from dataclasses import dataclass
from datetime import date as _date
from datetime import datetime as _datetime
from pathlib import Path, PurePath
from tempfile import mktemp
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
    Union,
)
from urllib.parse import urlparse, urlsplit, urlunsplit
from zipfile import ZipFile, is_zipfile

import pooch  # pyright: ignore [reportMissingTypeStubs]
import pydantic
import requests
from pydantic import (
    AnyUrl,
    DirectoryPath,
    Field,
    GetCoreSchemaHandler,
    PlainSerializer,
    PrivateAttr,
    RootModel,
    SerializationInfo,
    TypeAdapter,
    model_validator,
)
from pydantic_core import core_schema
from tqdm import tqdm
from typing_extensions import (
    Annotated,
    LiteralString,
    NotRequired,
    Self,
    TypeGuard,
    Unpack,
    assert_never,
)
from typing_extensions import TypeAliasType as _TypeAliasType

from ._settings import settings
from .io_basics import (
    ALL_BIOIMAGEIO_YAML_NAMES,
    ALTERNATIVE_BIOIMAGEIO_YAML_NAMES,
    BIOIMAGEIO_YAML,
    AbsoluteDirectory,
    AbsoluteFilePath,
    FileName,
    FilePath,
    Sha256,
    ZipPath,
)
from .node import Node
from .packaging_context import packaging_context_var
from .root_url import RootHttpUrl
from .type_guards import is_mapping, is_sequence
from .url import HttpUrl
from .validation_context import get_validation_context
from .validator_annotations import AfterValidator

if sys.version_info < (3, 10):
    SLOTS: Dict[str, bool] = {}
else:
    SLOTS = {"slots": True}


AbsolutePathT = TypeVar(
    "AbsolutePathT",
    bound=Union[HttpUrl, AbsoluteDirectory, AbsoluteFilePath, ZipPath],
)


class RelativePathBase(RootModel[PurePath], Generic[AbsolutePathT], frozen=True):
    _absolute: AbsolutePathT = PrivateAttr()

    @property
    def path(self) -> PurePath:
        return self.root

    def absolute(  # method not property analog to `pathlib.Path.absolute()`
        self,
    ) -> AbsolutePathT:
        """get the absolute path/url

        (resolved at time of initialization with the root of the ValidationContext)
        """
        return self._absolute

    def model_post_init(self, __context: Any) -> None:
        """set `_absolute` property with validation context at creation time. @private"""
        if self.root.is_absolute():
            raise ValueError(f"{self.root} is an absolute path.")

        if self.root.parts and self.root.parts[0] in ("http:", "https:"):
            raise ValueError(f"{self.root} looks like an http url.")

        self._absolute = (  # pyright: ignore[reportAttributeAccessIssue]
            self.get_absolute(get_validation_context().root)
        )
        super().model_post_init(__context)

    # @property
    # def __members(self):
    #     return (self.path,)

    # def __eq__(self, __value: object) -> bool:
    #     return type(__value) is type(self) and self.__members == __value.__members

    # def __hash__(self) -> int:
    #     return hash(self.__members)

    def __str__(self) -> str:
        return self.root.as_posix()

    def __repr__(self) -> str:
        return f"RelativePath('{self}')"

    @abstractmethod
    def get_absolute(
        self, root: Union[RootHttpUrl, AbsoluteDirectory, pydantic.AnyUrl, ZipFile]
    ) -> AbsolutePathT: ...

    def _get_absolute_impl(
        self, root: Union[RootHttpUrl, AbsoluteDirectory, pydantic.AnyUrl, ZipFile]
    ) -> Union[Path, HttpUrl, ZipPath]:
        if isinstance(root, Path):
            return (root / self.root).absolute()

        rel_path = self.root.as_posix().strip("/")
        if isinstance(root, ZipFile):
            return ZipPath(root, rel_path)

        parsed = urlsplit(str(root))
        path = list(parsed.path.strip("/").split("/"))
        if (
            parsed.netloc == "zenodo.org"
            and parsed.path.startswith("/api/records/")
            and parsed.path.endswith("/content")
        ):
            path.insert(-1, rel_path)
        else:
            path.append(rel_path)

        return HttpUrl(
            urlunsplit(
                (
                    parsed.scheme,
                    parsed.netloc,
                    "/".join(path),
                    parsed.query,
                    parsed.fragment,
                )
            )
        )

    @classmethod
    def _validate(cls, value: Union[PurePath, str]):
        if isinstance(value, str) and (
            value.startswith("https://") or value.startswith("http://")
        ):
            raise ValueError(f"{value} looks like a URL, not a relative path")

        return cls(PurePath(value))


class RelativeFilePath(
    RelativePathBase[Union[AbsoluteFilePath, HttpUrl, ZipPath]], frozen=True
):
    """A path relative to the `rdf.yaml` file (also if the RDF source is a URL)."""

    def model_post_init(self, __context: Any) -> None:
        """add validation @private"""
        if not self.root.parts:  # an empty path can only be a directory
            raise ValueError(f"{self.root} is not a valid file path.")

        super().model_post_init(__context)

    def get_absolute(
        self, root: "RootHttpUrl | Path | AnyUrl | ZipFile"
    ) -> "AbsoluteFilePath | HttpUrl | ZipPath":
        absolute = self._get_absolute_impl(root)
        if (
            isinstance(absolute, Path)
            and (context := get_validation_context()).perform_io_checks
            and str(self.root) not in context.known_files
            and not absolute.is_file()
        ):
            raise ValueError(f"{absolute} does not point to an existing file")

        return absolute


class RelativeDirectory(
    RelativePathBase[Union[AbsoluteDirectory, HttpUrl, ZipPath]], frozen=True
):
    def get_absolute(
        self, root: "RootHttpUrl | Path | AnyUrl | ZipFile"
    ) -> "AbsoluteDirectory | HttpUrl | ZipPath":
        absolute = self._get_absolute_impl(root)
        if (
            isinstance(absolute, Path)
            and get_validation_context().perform_io_checks
            and not absolute.is_dir()
        ):
            raise ValueError(f"{absolute} does not point to an existing directory")

        return absolute


FileSource = Annotated[
    Union[HttpUrl, RelativeFilePath, FilePath],
    Field(union_mode="left_to_right"),
]
PermissiveFileSource = Union[FileSource, str, pydantic.HttpUrl]

V_suffix = TypeVar("V_suffix", bound=FileSource)
path_or_url_adapter: "TypeAdapter[Union[FilePath, DirectoryPath, HttpUrl]]" = (
    TypeAdapter(Union[FilePath, DirectoryPath, HttpUrl])
)


def validate_suffix(
    value: V_suffix, suffix: Union[str, Sequence[str]], case_sensitive: bool
) -> V_suffix:
    """check final suffix"""
    if isinstance(suffix, str):
        suffixes = [suffix]
    else:
        suffixes = suffix

    assert len(suffixes) > 0, "no suffix given"
    assert all(
        suff.startswith(".") for suff in suffixes
    ), "expected suffixes to start with '.'"
    o_value = value
    strict = interprete_file_source(value)

    if isinstance(strict, (HttpUrl, AnyUrl)):
        if strict.path is None or "." not in (path := strict.path):
            actual_suffix = ""
        elif (
            strict.host == "zenodo.org"
            and path.startswith("/api/records/")
            and path.endswith("/content")
        ):
            actual_suffix = "." + path[: -len("/content")].split(".")[-1]
        else:
            actual_suffix = "." + path.split(".")[-1]

    elif isinstance(strict, PurePath):
        actual_suffix = strict.suffixes[-1]
    elif isinstance(strict, RelativeFilePath):
        actual_suffix = strict.path.suffixes[-1]
    else:
        assert_never(strict)

    if (
        case_sensitive
        and actual_suffix not in suffixes
        or not case_sensitive
        and actual_suffix.lower() not in [s.lower() for s in suffixes]
    ):
        if len(suffixes) == 1:
            raise ValueError(f"Expected suffix {suffixes[0]}, but got {actual_suffix}")
        else:
            raise ValueError(
                f"Expected a suffix from {suffixes}, but got {actual_suffix}"
            )

    return o_value


@dataclass(frozen=True, **SLOTS)
class WithSuffix:
    suffix: Union[LiteralString, Tuple[LiteralString, ...]]
    case_sensitive: bool

    def __get_pydantic_core_schema__(
        self, source: Type[Any], handler: GetCoreSchemaHandler
    ):
        if not self.suffix:
            raise ValueError("suffix may not be empty")

        schema = handler(source)
        return core_schema.no_info_after_validator_function(
            self.validate,
            schema,
        )

    def validate(self, value: FileSource) -> FileSource:
        return validate_suffix(value, self.suffix, case_sensitive=self.case_sensitive)


def wo_special_file_name(src: FileSource) -> FileSource:
    if has_valid_bioimageio_yaml_name(src):
        raise ValueError(
            f"'{src}' not allowed here as its filename is reserved to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return src


def _package(value: FileSource, info: SerializationInfo) -> Union[str, Path, FileName]:
    if (packaging_context := packaging_context_var.get()) is None:
        # convert to standard python obj
        # note: pydantic keeps returning Rootmodels (here `HttpUrl`) as-is, but if
        #   this function returns one RootModel, paths are "further serialized" by
        #   returning the 'root' attribute, which is incorrect.
        #   see https://github.com/pydantic/pydantic/issues/8963
        #   TODO: follow up on https://github.com/pydantic/pydantic/issues/8963
        if isinstance(value, Path):
            unpackaged = value
        elif isinstance(value, HttpUrl):
            unpackaged = value
        elif isinstance(value, RelativeFilePath):
            unpackaged = Path(value.path)
        elif isinstance(value, AnyUrl):
            unpackaged = str(value)
        else:
            assert_never(value)

        if info.mode_is_json():
            # convert to json value  # TODO: remove and let pydantic do this?
            if isinstance(unpackaged, Path):
                unpackaged = str(unpackaged)
            elif isinstance(unpackaged, str):
                pass
            else:
                assert_never(unpackaged)
        else:
            warnings.warn(
                "dumping with mode='python' is currently not fully supported for "
                + "fields that are included when packaging; returned objects are "
                + "standard python objects"
            )

        return unpackaged  # return unpackaged file source

    # package the file source:
    # add it to the current package's file sources and return its collision free file name
    if isinstance(value, RelativeFilePath):
        src = value.absolute()
    elif isinstance(value, pydantic.AnyUrl):
        src = HttpUrl(str(value))
    elif isinstance(value, HttpUrl):
        src = value
    elif isinstance(value, Path):
        src = value.resolve()
    else:
        assert_never(value)

    fname = extract_file_name(src)
    if fname == packaging_context.bioimageio_yaml_file_name:
        raise ValueError(
            f"Reserved file name '{packaging_context.bioimageio_yaml_file_name}' "
            + "not allowed for a file to be packaged"
        )

    fsrcs = packaging_context.file_sources
    assert not any(
        fname.endswith(special) for special in ALL_BIOIMAGEIO_YAML_NAMES
    ), fname
    if fname in fsrcs and fsrcs[fname] != src:
        for i in range(2, 20):
            fn, *ext = fname.split(".")
            alternative_file_name = ".".join([f"{fn}_{i}", *ext])
            if (
                alternative_file_name not in fsrcs
                or fsrcs[alternative_file_name] == src
            ):
                fname = alternative_file_name
                break
        else:
            raise ValueError(f"Too many file name clashes for {fname}")

    fsrcs[fname] = src
    return fname


include_in_package_serializer = PlainSerializer(_package, when_used="unless-none")
ImportantFileSource = Annotated[
    FileSource,
    AfterValidator(wo_special_file_name),
    include_in_package_serializer,
]
InPackageIfLocalFileSource = Union[
    Annotated[
        Union[FilePath, RelativeFilePath],
        AfterValidator(wo_special_file_name),
        include_in_package_serializer,
    ],
    Union[HttpUrl, pydantic.HttpUrl],
]


def has_valid_bioimageio_yaml_name(src: FileSource) -> bool:
    return is_valid_bioimageio_yaml_name(extract_file_name(src))


def is_valid_bioimageio_yaml_name(file_name: FileName) -> bool:
    for bioimageio_name in ALL_BIOIMAGEIO_YAML_NAMES:
        if file_name == bioimageio_name or file_name.endswith("." + bioimageio_name):
            return True

    return False


def identify_bioimageio_yaml_file_name(file_names: Iterable[FileName]) -> FileName:
    file_names = sorted(file_names)
    for bioimageio_name in ALL_BIOIMAGEIO_YAML_NAMES:
        for file_name in file_names:
            if file_name == bioimageio_name or file_name.endswith(
                "." + bioimageio_name
            ):
                return file_name

    raise ValueError(
        f"No {BIOIMAGEIO_YAML} found in {file_names}. (Looking for '{BIOIMAGEIO_YAML}'"
        + " or or any of the alterntive file names:"
        + f" {ALTERNATIVE_BIOIMAGEIO_YAML_NAMES}, or any file with an extension of"
        + f"  those, e.g. 'anything.{BIOIMAGEIO_YAML}')."
    )


def find_bioimageio_yaml_file_name(path: Union[Path, ZipFile]) -> FileName:
    if isinstance(path, ZipFile):
        file_names = path.namelist()
    elif path.is_file():
        if not is_zipfile(path):
            return path.name

        with ZipFile(path, "r") as f:
            file_names = f.namelist()
    else:
        file_names = [p.name for p in path.glob("*")]

    return identify_bioimageio_yaml_file_name(file_names)


def ensure_has_valid_bioimageio_yaml_name(src: FileSource) -> FileSource:
    if not has_valid_bioimageio_yaml_name(src):
        raise ValueError(
            f"'{src}' does not have a valid filename to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return src


def ensure_is_valid_bioimageio_yaml_name(file_name: FileName) -> FileName:
    if not is_valid_bioimageio_yaml_name(file_name):
        raise ValueError(
            f"'{file_name}' is not a valid filename to identify"
            + f" '{BIOIMAGEIO_YAML}' (or equivalent) files."
        )

    return file_name


# types as loaded from YAML 1.2 (with ruyaml)
YamlLeafValue = Union[
    bool, _date, _datetime, int, float, str, None
]  # note: order relevant for deserializing
YamlKey = Union[  # YAML Arrays are cast to tuples if used as key in mappings
    YamlLeafValue, Tuple[YamlLeafValue, ...]  # (nesting is not allowed though)
]
if TYPE_CHECKING:
    YamlValue = Union[YamlLeafValue, List["YamlValue"], Dict[YamlKey, "YamlValue"]]
else:
    # for pydantic validation we need to use `TypeAliasType`,
    # see https://docs.pydantic.dev/latest/concepts/types/#named-recursive-types
    # however this results in a partially unknown type with the current pyright 1.1.388
    YamlValue = _TypeAliasType(
        "YamlValue",
        Union[YamlLeafValue, List["YamlValue"], Dict[YamlKey, "YamlValue"]],
    )
BioimageioYamlContent = Dict[str, YamlValue]
BioimageioYamlSource = Union[PermissiveFileSource, ZipFile, BioimageioYamlContent]


def is_yaml_leaf_value(value: Any) -> TypeGuard[YamlLeafValue]:
    return isinstance(value, (bool, _date, _datetime, int, float, str, type(None)))


def is_yaml_list(value: Any) -> TypeGuard[List[YamlValue]]:
    return is_sequence(value) and all(is_yaml_value(item) for item in value)


def is_yaml_mapping(value: Any) -> TypeGuard[BioimageioYamlContent]:
    return is_mapping(value) and all(
        isinstance(key, str) and is_yaml_value(val) for key, val in value.items()
    )


def is_yaml_value(value: Any) -> TypeGuard[YamlValue]:
    return is_yaml_leaf_value(value) or is_yaml_list(value) or is_yaml_mapping(value)


@dataclass
class OpenedBioimageioYaml:
    content: BioimageioYamlContent
    original_root: Union[AbsoluteDirectory, RootHttpUrl, ZipFile]
    original_file_name: FileName
    unparsed_content: str


@dataclass
class LocalFile:
    path: FilePath
    original_root: Union[AbsoluteDirectory, RootHttpUrl, ZipFile]
    original_file_name: FileName


@dataclass
class FileInZip:
    path: ZipPath
    original_root: Union[RootHttpUrl, ZipFile]
    original_file_name: FileName


class HashKwargs(TypedDict):
    sha256: NotRequired[Optional[Sha256]]


_file_source_adapter: TypeAdapter[Union[HttpUrl, RelativeFilePath, FilePath]] = (
    TypeAdapter(FileSource)
)


def interprete_file_source(file_source: PermissiveFileSource) -> FileSource:
    if isinstance(file_source, Path):
        if file_source.is_dir():
            raise FileNotFoundError(
                f"{file_source} is a directory, but expected a file."
            )
        return file_source

    if isinstance(file_source, HttpUrl):
        return file_source

    if isinstance(file_source, pydantic.AnyUrl):
        file_source = str(file_source)

    with get_validation_context().replace(perform_io_checks=False):
        strict = _file_source_adapter.validate_python(file_source)
        if isinstance(strict, Path) and strict.is_dir():
            raise FileNotFoundError(f"{strict} is a directory, but expected a file.")

    return strict


def _get_known_hash(hash_kwargs: HashKwargs):
    if "sha256" in hash_kwargs and hash_kwargs["sha256"] is not None:
        return f"sha256:{hash_kwargs['sha256']}"
    else:
        return None


def _get_unique_file_name(url: Union[HttpUrl, pydantic.HttpUrl]):
    """
    Create a unique file name based on the given URL;
    adapted from pooch.utils.unique_file_name
    """
    md5 = hashlib.md5(str(url).encode()).hexdigest()
    fname = extract_file_name(url)
    # Crop the start of the file name to fit 255 characters including the hash
    # and the :
    fname = fname[-(255 - len(md5) - 1) :]
    unique_name = f"{md5}-{fname}"
    return unique_name


class Progressbar(Protocol):
    count: int
    total: int

    def update(self, i: int): ...

    def reset(self): ...

    def close(self): ...


def extract(
    source: Union[FilePath, ZipFile, ZipPath],
    folder: Optional[DirectoryPath] = None,
    overwrite: bool = False,
) -> DirectoryPath:
    extract_member = None
    if isinstance(source, ZipPath):
        extract_member = source.at
        source = source.root

    if isinstance(source, ZipFile):
        zip_context = nullcontext(source)
        if folder is None:
            if source.filename is None:
                folder = Path(mktemp())
            else:
                zip_path = Path(source.filename)
                folder = zip_path.with_suffix(zip_path.suffix + ".unzip")
    else:
        zip_context = ZipFile(source, "r")
        if folder is None:
            folder = source.with_suffix(source.suffix + ".unzip")

    if overwrite and folder.exists():
        warnings.warn(f"Overwriting existing unzipped archive at {folder}")

    with zip_context as f:
        if extract_member is not None:
            extracted_file_path = folder / extract_member
            if extracted_file_path.exists() and not overwrite:
                warnings.warn(f"Found unzipped {extracted_file_path}.")
            else:
                _ = f.extract(extract_member, folder)

            return folder

        elif overwrite or not folder.exists():
            f.extractall(folder)
            return folder

        found_content = {p.relative_to(folder).as_posix() for p in folder.glob("*")}
        expected_content = {info.filename for info in f.filelist}
        if expected_missing := expected_content - found_content:
            parts = folder.name.split("_")
            nr, *suffixes = parts[-1].split(".")
            if nr.isdecimal():
                nr = str(int(nr) + 1)
            else:
                nr = f"1.{nr}"

            parts[-1] = ".".join([nr, *suffixes])
            out_path_new = folder.with_name("_".join(parts))
            warnings.warn(
                f"Unzipped archive at {folder} is missing expected files"
                + f" {expected_missing}."
                + f" Unzipping to {out_path_new} instead to avoid overwriting."
            )
            return extract(f, out_path_new, overwrite=overwrite)
        else:
            warnings.warn(
                f"Found unzipped archive with all expected files at {folder}."
            )
            return folder


def resolve(
    source: Union[PermissiveFileSource, FileDescr, ZipPath],
    /,
    progressbar: Union[Progressbar, bool, None] = None,
    **kwargs: Unpack[HashKwargs],
) -> Union[LocalFile, FileInZip]:
    """Resolve file `source` (download if needed)"""

    if isinstance(source, str):
        source = interprete_file_source(source)

    if isinstance(source, RelativeFilePath):
        source = source.absolute()
        if isinstance(source, ZipPath):
            return FileInZip(source, source.root, extract_file_name(source))

    if isinstance(source, pydantic.AnyUrl):
        with get_validation_context().replace(perform_io_checks=False):
            source = HttpUrl(source)

    if isinstance(source, FileDescr):
        return source.download()
    elif isinstance(source, ZipPath):
        zip_root = source.root
        assert isinstance(zip_root, ZipFile)
        return FileInZip(
            source,
            zip_root,
            extract_file_name(source),
        )
    elif isinstance(source, Path):
        if source.is_dir():
            raise FileNotFoundError(f"{source} is a directory, not a file")

        if not source.exists():
            raise FileNotFoundError(source)

        return LocalFile(
            source,
            source.parent,
            extract_file_name(source),
        )
    elif isinstance(source, HttpUrl):
        if source.scheme not in ("http", "https"):
            raise NotImplementedError(source.scheme)

        if settings.CI:
            headers = {"User-Agent": "ci"}
            if progressbar is None:
                progressbar = False
        else:
            headers = {}
            if progressbar is None:
                progressbar = True

        if settings.user_agent is not None:
            headers["User-Agent"] = settings.user_agent

        chunk_size = 1024
        if (
            settings.cache_path
            and not get_validation_context().disable_cache
            and any(v is not None for v in kwargs.values())
        ):
            downloader = pooch.HTTPDownloader(
                headers=headers,
                progressbar=progressbar,  # pyright: ignore[reportArgumentType]
                chunk_size=chunk_size,
            )
            fname = _get_unique_file_name(source)
            _ls: Any = pooch.retrieve(
                url=str(source),
                known_hash=_get_known_hash(kwargs),
                downloader=downloader,
                fname=fname,
                path=settings.cache_path,
            )
            local_source = Path(_ls).absolute()
            return LocalFile(
                local_source,
                source.parent,
                extract_file_name(source),
            )
        else:
            # cacheless download to memory using an in memory zip file
            r = requests.get(str(source), stream=True)
            r.raise_for_status()

            zf = zipfile.ZipFile(io.BytesIO(), "w")
            fn = extract_file_name(source)
            total = int(r.headers.get("content-length", 0))

            if isinstance(progressbar, bool):
                if progressbar:
                    use_ascii = bool(sys.platform == "win32")
                    pbar = tqdm(
                        total=total,
                        ncols=79,
                        ascii=use_ascii,
                        unit="B",
                        unit_scale=True,
                        leave=True,
                    )
                    pbar = tqdm(desc=f"Downloading {fn}")
                else:
                    pbar = None
            else:
                pbar = progressbar

            zp = ZipPath(zf, fn)
            with zp.open("wb") as z:
                assert not isinstance(z, io.TextIOWrapper)
                for chunk in r.iter_content(chunk_size=chunk_size):
                    n = z.write(chunk)
                    if pbar is not None:
                        _ = pbar.update(n)

            # Make sure the progress bar gets filled even if the actual number
            # is chunks is smaller than expected. This happens when streaming
            # text files that are compressed by the server when sending (gzip).
            # Binary files don't experience this.
            # (adapted from pooch.HttpDownloader)
            if pbar is not None:
                pbar.reset()
                _ = pbar.update(total)
                pbar.close()

            return FileInZip(
                path=zp,
                original_root=source.parent,
                original_file_name=fn,
            )

    else:
        assert_never(source)


download = resolve


def resolve_and_extract(
    source: Union[PermissiveFileSource, FileDescr, ZipPath],
    /,
    progressbar: Union[Progressbar, bool, None] = None,
    **kwargs: Unpack[HashKwargs],
) -> LocalFile:
    """Resolve `source` within current ValidationContext,
    download if needed and
    extract file if within zip archive.

    note: If source points to a zip file it is not extracted
    """
    local = resolve(source, progressbar=progressbar, **kwargs)
    if isinstance(local, LocalFile):
        return local

    folder = extract(local.path)

    return LocalFile(
        folder / local.path.at,
        original_root=local.original_root,
        original_file_name=local.original_file_name,
    )


class LightHttpFileDescr(Node):
    """http source with sha256 value (minimal validation)"""

    source: pydantic.HttpUrl
    """file source"""

    sha256: Sha256
    """SHA256 checksum of the source file"""

    def download(self):
        return download(self.source, sha256=self.sha256)


class FileDescr(Node):
    source: ImportantFileSource
    """∈📦 file source"""

    sha256: Optional[Sha256] = None
    """SHA256 checksum of the source file"""

    @model_validator(mode="after")
    def _validate_sha256(self) -> Self:
        if get_validation_context().perform_io_checks:
            self.validate_sha256()

        return self

    def validate_sha256(self):
        context = get_validation_context()
        if (src_str := str(self.source)) in context.known_files:
            actual_sha = context.known_files[src_str]
        else:
            local_source = download(self.source, sha256=self.sha256).path
            actual_sha = get_sha256(local_source)
            context.known_files[src_str] = actual_sha

        if actual_sha is None:
            return
        elif self.sha256 == actual_sha:
            pass
        elif self.sha256 is None or context.update_hashes:
            self.sha256 = actual_sha
        elif self.sha256 != actual_sha:
            raise ValueError(
                f"Sha256 mismatch for {self.source}. Expected {self.sha256}, got "
                + f"{actual_sha}. Update expected `sha256` or point to the matching "
                + "file."
            )

    def download(self):

        return download(self.source, sha256=self.sha256)


def extract_file_name(
    src: Union[pydantic.HttpUrl, HttpUrl, PurePath, RelativeFilePath, ZipPath],
) -> FileName:
    if isinstance(src, ZipPath):
        return src.name or src.root.filename or "bioimageio.zip"
    elif isinstance(src, RelativeFilePath):
        return src.path.name
    elif isinstance(src, PurePath):
        return src.name
    else:
        url = urlparse(str(src))
        if (
            url.scheme == "https"
            and url.hostname == "zenodo.org"
            and url.path.startswith("/api/records/")
            and url.path.endswith("/content")
        ):
            return url.path.split("/")[-2]
        else:
            return url.path.split("/")[-1]


def get_sha256(path: Union[Path, ZipPath]) -> Sha256:
    """from https://stackoverflow.com/a/44873382"""
    if isinstance(path, ZipPath):
        # no buffered reading available
        zf = path.root
        assert isinstance(zf, ZipFile)
        data = path.read_bytes()
        assert isinstance(data, bytes)
        h = hashlib.sha256(data)
    else:
        h = hashlib.sha256()
        chunksize = 128 * 1024
        b = bytearray(chunksize)
        mv = memoryview(b)
        with open(path, "rb", buffering=0) as f:
            for n in iter(lambda: f.readinto(mv), 0):
                h.update(mv[:n])

    sha = h.hexdigest()
    assert len(sha) == 64
    return Sha256(sha)
